import argparse
import importlib
import os
import subprocess
import sys
import uuid
from urllib.parse import urlparse

from pyspark.sql import SparkSession, SQLContext, Row

# lib installer
def _install_libs(lib_url):
    current_dir = os.getcwd()
    unique_id = str(uuid.uuid4())
    lib_dir = os.path.join(current_dir, unique_id, 'python_libs')
    lib_zip = os.path.join(current_dir, unique_id, 'lib.zip')
    
    os.makedirs(lib_dir)
    subprocess.check_call(['wget', "-O", lib_zip, lib_url])
    subprocess.check_call(['unzip', lib_zip, "-d", lib_dir])
    sys.path.insert(0, lib_dir)

def _save_result(os_client, run_dir, result):
    from oci_core import os_upload_json
    o = urlparse(run_dir)

    namespace = o.netloc.split('@')[1]
    bucket = o.netloc.split('@')[0]
    object_name_prefix = o.path[1:]  # drop the leading /

    os_upload_json(os_client, result, namespace, bucket, f"{object_name_prefix}/result.json")

def _get_args(os_client, run_dir):
    from oci_core import os_download_json
    o = urlparse(run_dir)

    namespace = o.netloc.split('@')[1]
    bucket = o.netloc.split('@')[0]
    object_name_prefix = o.path[1:]  # drop the leading /

    return os_download_json(os_client, namespace, bucket, f"{object_name_prefix}/args.json")


def _bootstrap():
    parser = argparse.ArgumentParser(description='job')
    parser.add_argument(
        "--deployment-location", type=str, required=True, help="Deployment Location",
    )
    parser.add_argument(
        "--run-id", type=str, required=True, help="Run ID",
    )
    parser.add_argument(
        "--run-dir", type=str, required=True, help="Run Directory",
    )
    parser.add_argument(
        "--lib-url", type=str, required=True, help="Library URL",
    )
    parser.add_argument(
        "--app-region", type=str, required=True, help="Application Region",
    )
    args = parser.parse_args()

    o = urlparse(args.run_dir)
    if o.scheme != 'oci':
        raise Exception("run_dir MUST be in OCI")

    # Load application
    spark = SparkSession.builder.appName("RunJob").getOrCreate()
    sc = spark.sparkContext
    sc.addPyFile(f"{args.deployment_location}/app.zip")
    print("Application loaded")

    subprocess.check_call(['wget', args.lib_url])

    # The archive file goes into /opt/dataflow
    # Load python libraries
    current_dir = os.getcwd()
    lib_dir = os.path.join(current_dir, 'python_libs')
    os.mkdir(lib_dir)
    subprocess.call([
        'unzip', "lib.zip", "-d", lib_dir
    ])
    sys.path.insert(0, lib_dir)

    run_id = args.run_id
    run_dir = args.run_dir

    # The entry is in a python file "main.py"
    from oci_core import dfapp_get_os_client, get_delegation_token
    delegation_token = get_delegation_token(spark)
    os_client = dfapp_get_os_client(args.app_region, delegation_token)
    xargs = _get_args(os_client, run_dir)
    entry = importlib.import_module("main")
    result = entry.main(spark, xargs, sysops={
        "install_libs": lambda : _install_libs(args.lib_url)
    })

    _save_result(os_client, run_dir, result)

    # TODO: make a copy of the log file


_bootstrap()



import importlib
import argparse
import argparse
import uuid
import os
import subprocess
import sys
import tempfile
import json

from pyspark.sql import SparkSession

# lib installer
def _install_libs(lib_url):
    ##########################################
    #
    # |
    # +-- python_libs
    #       |
    #       +-- uuid1.zip
    #       |
    #       +-- uuid2.zip
    #       |
    #       +-- uuid1  (lib extracted)
    #       |
    #       +-- uuid2  (lib extracted)
    ##########################################
    current_dir = os.getcwd()
    unique_id = str(uuid.uuid4())
    lib_dir = os.path.join(current_dir, 'python_libs')
    bin_dir = os.path.join(lib_dir, unique_id)
    lib_zip = os.path.join(lib_dir, f"{unique_id}.zip")
    os.makedirs(bin_dir, exist_ok=True)

    # lib_url must be a HDFS url (or perhaps can be handled by a hdfs connector)    
    subprocess.check_call(["hdfs", "dfs", "-copyToLocal", lib_url, lib_zip])
    subprocess.check_call(['unzip', lib_zip, "-d", bin_dir])
    sys.path.insert(0, bin_dir)


print("job_loader.py: enter")

parser = argparse.ArgumentParser(description='job')
parser.add_argument(
    "--run-id", type=str, required=True, help="Run ID",
)
parser.add_argument(
    "--run-dir", type=str, required=True, help="Run Directory",
)
parser.add_argument(
    "--lib-url", type=str, required=True, help="Library URL",
)

args = parser.parse_args()

_install_libs(args.lib_url)

# get input
input_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
input_file.close()
subprocess.check_call(["hdfs", "dfs", "-copyToLocal", "-f", f"{args.run_dir}/input.json", input_file.name])
with open(input_file.name) as f:
    input_args = json.load(f)

spark = SparkSession.builder.appName("RunJob").getOrCreate()

try:
    entry = importlib.import_module("main")
    result = entry.main(spark, input_args)

    # save output
    output_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
    json.dump(result, output_file)
    output_file.close()

    subprocess.check_call(["hdfs", "dfs", "-copyFromLocal", output_file.name, f"{args.run_dir}/result.json"])
    print("job_loader: exit gracefully")
finally:
    spark.stop()



import json
from urllib.parse import urlparse
import time
import uuid
from datetime import datetime, timedelta

import oci
from oci_core import get_os_client, get_df_client, os_upload, os_upload_json, os_download, os_download_json, os_get_endpoint

from spark_etl.job_submitters import AbstractJobSubmitter
from spark_etl import SparkETLLaunchFailure, SparkETLGetStatusFailure, SparkETLKillFailure
from .tools import check_response, remote_execute

class DataflowJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(DataflowJobSubmitter, self).__init__(config)
        # config fields
        # region, e.g. IAD
        # run_base_dir, uri, point to the run directory.
        run_base_dir = self.config['run_base_dir']
        o = urlparse(run_base_dir)
        if o.scheme != 'oci':
            raise SparkETLLaunchFailure("run_base_dir must be in OCI")

    
    @property
    def region(self):
        return self.config['region']

    def run(self, deployment_location, options={}, args={}):
        o = urlparse(deployment_location)
        if o.scheme != 'oci':
            raise SparkETLLaunchFailure("deployment_location must be in OCI")

        run_base_dir = self.config['run_base_dir']
        run_id = str(uuid.uuid4())

        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:] # remove the leading "/"

        # let's get the deployment.json
        os_client = get_os_client(self.region)
        deployment = os_download_json(os_client, namespace, bucket, f"{root_path}/deployment.json")

        r = os_client.create_preauthenticated_request(
            namespace,
            bucket,
            oci.object_storage.models.CreatePreauthenticatedRequestDetails(
                access_type = 'ObjectRead',
                name=f'for run {run_id}',
                object_name=f'{root_path}/lib.zip',
                time_expires=datetime.utcnow() + timedelta(minutes=30)
            )
        )
        check_response(r, lambda : SparkETLLaunchFailure("dataflow failed to get lib url"))
        lib_url = f"{os_get_endpoint(self.region)}{r.data.access_uri}"

        # let's upload the args
        o = urlparse(self.config['run_base_dir'])
        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:] # remove the leading "/"
        os_upload_json(os_client, args, namespace, bucket, f"{root_path}/{run_id}/args.json")

        df_client = get_df_client(self.region)
        crd_argv = {
            'compartment_id': deployment['compartment_id'],
            'application_id': deployment['application_id'],
            'display_name' :options["display_name"],
            'arguments': [
                "--deployment-location", deployment_location,
                "--run-id", run_id,
                "--run-dir", f"{run_base_dir}/{run_id}",
                "--app-region", self.region,
                "--lib-url", lib_url,
            ],
        }
        for key in ['num_executors', 'driver_shape', 'executor_shape']:
            if key in options:
                crd_argv[key] = options[key]

        create_run_details = oci.data_flow.models.CreateRunDetails(**crd_argv)
        r = df_client.create_run(create_run_details=create_run_details)
        check_response(r, lambda : SparkETLLaunchFailure("dataflow failed to run the application"))
        run = r.data
        run_id = run.id
        print(f"Job launched, run_id = {run_id}")

        while True:
            time.sleep(10)
            r = df_client.get_run(run_id=run.id)
            check_response(r, lambda : SparkETLGetStatusFailure("dataflow failed to get run status"))
            run = r.data
            print(f"Status: {run.lifecycle_state}")
            if run.lifecycle_state in ('FAILED', 'SUCCEEDED'):
                break
        
        return {
            'state': run.lifecycle_state,
            'run_id': run_id,
            'succeeded': run.lifecycle_state == 'SUCCEEDED'
        }

    def get_result(self, run_id):
        result_object_name = f"{self.config['run_base_dir']}/{run_id}/result.json"
        o = urlparse(result_object_name)

        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        object_name = o.path[1:]    # remove the leading "/"

        os_client = get_os_client(self.region)
        result = os_download_json(os_client, namespace, bucket, object_name)
        return result


    def kill(self, run_id):
        df_client = get_df_client(self.region)

        r = df_client.delete_run(run_id)
        check_response(r, lambda : SparkETLKillFailure("dataflow failed to kill the run"))



import json

class Build:
    # This represent an application build
    def __init__(self, build_dir):
        self.build_dir = build_dir
        with open(f"{self.build_dir}/manifest.json", "r") as f:
            self.manifest = json.load(f)
    
    @property
    def artifacts(self):
        return ("app.zip", "lib.zip", "main.py", "manifest.json")

    @property
    def version(self):
        return self.manifest["version"]



from .application import Application
from .build import Build
from .exceptions import SparkETLException, SparkETLDeploymentFailure, SparkETLLaunchFailure, SparkETLGetStatusFailure, \
    SparkETLKillFailure



import os
import json
import time
import subprocess
from urllib.parse import urlparse
import uuid
import tempfile

from requests.auth import HTTPBasicAuth
import requests

from .abstract_job_submitter import AbstractJobSubmitter

def _execute(host, cmd, error_ok=False):
    r = subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
    if not error_ok and r != 0:
        raise Exception(f"command {cmd} failed with exit code {r}")

class LivyJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(LivyJobSubmitter, self).__init__(config)

    # options is vendor specific arguments
    # args.conf is the spark job config
    # args is arguments need to pass to the main entry
    def run(self, deployment_location, options={}, args={}):
        bridge = self.config["bridge"]
        stage_dir = self.config['stage_dir']
        run_dir = self.config['run_dir']

        # create run dir
        run_id = str(uuid.uuid4())
        _execute(bridge, f"mkdir -p {stage_dir}/{run_id}")
        _execute(bridge, f"hdfs dfs -mkdir -p {run_dir}/{run_id}")

        # generate input.json
        tf = tempfile.NamedTemporaryFile(mode="wt", delete=False)
        json.dump(args, tf)
        tf.close()
        subprocess.check_call([
            'scp', '-q', tf.name, f"{bridge}:{stage_dir}/{run_id}/input.json"
        ])
        _execute(bridge, f"hdfs dfs -copyFromLocal {stage_dir}/{run_id}/input.json {run_dir}/{run_id}/input.json")
        os.remove(tf.name)

        o = urlparse(deployment_location)
        if o.scheme != 'hdfs':
            raise SparkETLDeploymentFailure("deployment_location must be in hdfs")

        headers = {
            "Content-Type": "application/json",
            "X-Requested-By": "root",
            'proxyUser': 'root'
        }

        config = {
            'file': f'{deployment_location}/job_loader.py',
            'pyFiles': [f'{deployment_location}/app.zip'],
            'args': [
                '--run-id', run_id, 
                '--run-dir', f"{run_dir}/{run_id}",
                '--lib-url', f'{deployment_location}/lib.zip'
            ]
        }
        config.update(options)

        service_url = self.config['service_url']
        if service_url.endswith("/"):
            service_url = service_url[:-1]

        username = self.config['username']
        password = self.config['password']

        print(json.dumps(config))
        r = requests.post(
            "{}/batches".format(service_url),
            data=json.dumps(config),
            headers=headers,
            auth=HTTPBasicAuth(username, password),
            verify=False
        )
        if r.status_code not in [200, 201]:
            msg = "Failed to submit the job, status: {}, error message: \"{}\"".format(
                r.status_code,
                r.content
            )
            raise Exception(msg)

        print('job submitted')
        ret = json.loads(r.content.decode("utf8"))
        job_id = ret['id']
        print('job id: {}'.format(job_id))
        print(ret)
        print('logs:')
        for log in ret.get('log', []):
            print(log)
        
        # pull the job status
        while True:
            r = requests.get(
                "{}/batches/{}".format(service_url, job_id),
                headers=headers,
                auth=HTTPBasicAuth(username, password),
                verify=False
            )
            if r.status_code != 200:
                msg = "Failed to get job status, status: {}, error message: \"{}\"".format(
                    r.status_code,
                    r.content
                )
                raise Exception(msg)
            ret = json.loads(r.content.decode("utf8"))
            job_state = ret['state']
            appId = ret.get('appId')
            print('job_state: {}, applicationId: {}'.format(
                job_state, appId

            ))
            if job_state in ['success', 'dead']:
                # cmd = f"yarn logs -applicationId {appId}"
                # host = self.config['bridge']
                # subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
                print(f"job_state={job_state}")
                if job_state == 'success':
                    # get the result.json
                    _execute(bridge, f"hdfs dfs -copyToLocal {run_dir}/{run_id}/result.json {stage_dir}/{run_id}/result.json")
                    rf = tempfile.NamedTemporaryFile(mode="w", delete=False)
                    rf.close()
                    subprocess.check_call([
                        'scp', '-q', f"{bridge}:{stage_dir}/{run_id}/result.json", rf.name
                    ])

                    with open(rf.name, "r") as f:
                        return json.load(f)
                raise Exception("Job failed")
            time.sleep(5)



class SparkETLException(Exception):
    def __init__(self, message, cause=None):
        super(SparkETLException, self).__init__(message)
        self.cause = cause

class SparkETLDeploymentFailure(SparkETLException):
    # Failed to deploy an application
    def __init__(self, message, cause=None):
        super(SparkETLDeploymentFailure, self).__init__(message, cause=cause)

class SparkETLLaunchFailure(SparkETLException):
    # Failed to launch an application
    def __init__(self, message, cause=None):
        super(SparkETLLaunchFailure, self).__init__(message, cause=cause)

class SparkETLKillFailure(SparkETLException):
    # Failed to kill an application run
    def __init__(self, message, cause=None):
        super(SparkETLKillFailure, self).__init__(message, cause=cause)

class SparkETLGetStatusFailure(SparkETLException):
    # Failed to get status of a launched application (aka run)
    def __init__(self, message, cause=None):
        super(SparkETLGetStatusFailure, self).__init__(message, cause=cause)



import subprocess

# Check a OCI api call response
def check_response(r, error_factory=None):
    if r.status != 200:
        if error_factory is None:
            error = Exception("request failed")
        else:
            error = error_factory()
        raise error

def remote_execute(host, cmd, error_ok=False):
    r = subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
    if not error_ok and r != 0:
        raise Exception(f"command \"{cmd}\" failed with exit code {r}")




import uuid
import subprocess
import os
from urllib.parse import urlparse

from .abstract_deployer import AbstractDeployer
from spark_etl import Build
from spark_etl.exceptions import SparkETLDeploymentFailure

def _execute(host, cmd, error_ok=False):
    r = subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
    if not error_ok and r != 0:
        raise Exception(f"command \"{cmd}\" failed with exit code {r}")


class HDFSDeployer(AbstractDeployer):
    """
    This deployer deploys application to HDFS
    """
    def __init__(self, config):
        super(HDFSDeployer, self).__init__(config)

    def deploy(self, build_dir, deployment_location):
        o = urlparse(deployment_location)
        if o.scheme != 'hdfs':
            raise SparkETLDeploymentFailure("deployment_location must be in hdfs")

        # let's copy files to the stage dir
        bridge_dir = os.path.join(self.config['stage_dir'], str(uuid.uuid4()))

        bridge = self.config["bridge"]
        _execute(bridge, f"mkdir -p {bridge_dir}")

        build = Build(build_dir)

        for artifact in build.artifacts:
            subprocess.call([
                'scp', '-q', f"{build_dir}/{artifact}", f"{bridge}:{bridge_dir}/{artifact}"
            ])

        # copy job loader
        job_loader_filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'job_loader.py'
        )
        subprocess.call([
            'scp', '-q', job_loader_filename, f"{bridge}:{bridge_dir}/job_loader.py"
        ])


        dest_location = f"{deployment_location}/{build.version}"
        _execute(bridge, f"hdfs dfs -rm -r {dest_location}", error_ok=True)
        _execute(bridge, f"hdfs dfs -mkdir -p {dest_location}")

        artifacts = []
        artifacts.extend(build.artifacts)
        artifacts.append("job_loader.py")
        for artifact in artifacts:
            _execute(bridge, f"hdfs dfs -copyFromLocal {bridge_dir}/{artifact} {dest_location}/{artifact}")
        
        _execute(bridge, f"rm -rf {bridge_dir}")



import uuid
import os
import subprocess
from urllib.parse import urlparse
import json
import tempfile

import oci
from oci_core import get_os_client, get_df_client, os_upload, os_upload_json

from spark_etl.deployers import AbstractDeployer
from spark_etl import Build, SparkETLDeploymentFailure
from .tools import check_response, remote_execute


class DataflowDeployer(AbstractDeployer):
    def __init__(self, config):
        super(DataflowDeployer, self).__init__(config)
        # config must have
        #   region, e.g., value is IAD
        #   compartment_id: the compartment we want to have the dataflow app
        #   driver_shape
        #   executor_shape
        #   num_executors
    
    @property
    def region(self):
        return self.config["region"]

    def create_application(self, manifest, destination_location):
        df_client = get_df_client(self.region)
        dataflow = self.config['dataflow']

        display_name = f"{manifest['display_name']}-{manifest['version']}"

        # if the application already exist, we will fail the deployment
        # since user should bump the version for new deployment
        r = df_client.list_applications(
            dataflow['compartment_id'],
            limit=1,
            display_name=display_name,
        )
        check_response(r, lambda : SparkETLDeploymentFailure("Unable to list application"))
        if len(r.data) > 0:
            raise SparkETLDeploymentFailure(f"Application {display_name} already created")

        create_application_details = oci.data_flow.models.CreateApplicationDetails(
            compartment_id=dataflow['compartment_id'],
            display_name=display_name,
            driver_shape=dataflow['driver_shape'],
            executor_shape=dataflow['executor_shape'],
            num_executors=dataflow['num_executors'],
            spark_version="2.4.4",
            file_uri=f"{destination_location}/{manifest['version']}/job_loader.py",
            language="PYTHON",
        )

        r = df_client.create_application(
            create_application_details=create_application_details
        )
        check_response(r, lambda : SparkETLDeploymentFailure("Unable to create dataflow application"))
        return r.data

    def deploy(self, build_dir, destination_location):
        o = urlparse(destination_location)
        if o.scheme != 'oci':
            raise SparkETLDeploymentFailure("destination_location must be in OCI")
        
        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:]    # remove the leading "/"

        build = Build(build_dir)

        print("Uploading files:")
        # Data flow want to call python lib python.zip
        os_client = get_os_client(self.region)
        for artifact in build.artifacts:
            os_upload(
                os_client, 
                f"{build_dir}/{artifact}", 
                namespace, 
                bucket, 
                f"{root_path}/{build.version}/{artifact}"
            )
        
        # let's upload the job loader
        job_loader_filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'job_loader.py'
        )
        os_upload(
            os_client, 
            job_loader_filename, 
            namespace, 
            bucket, 
            f"{root_path}/{build.version}/job_loader.py"
        )

        application = self.create_application(build.manifest, destination_location)
        app_info = {
            "application_id": application.id,
            "compartment_id": application.compartment_id
        }

        os_upload_json(
            os_client, app_info,
            namespace, bucket, f"{root_path}/{build.version}/deployment.json"
        )
